//
//  SSHelper.java
//  SwitchableString52
//
//  Created by Paul Suh on Fri Jul 25 2003.
/*
 
 Copyright (c) 2003 Paul Suh. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 
 * Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 
 * Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 
 * Neither the name of the site goodeast.com nor the names of its
 contributors may be used to endorse or promote products derived from
 this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 */

package com.goodeast.webobjects.switchablestring;


import com.webobjects.foundation.*;
import com.webobjects.appserver.*;
import java.util.*;
import java.text.*;


/** 
 * A helper class that is used to add behavior to the dynamic elements. This
 * class allows us to add multiple inheritance-like extensions to a class's
 * behavior within the constraints of Java. It has alternative implementations
 * of appendToResponse and takeValuesFromRequest that can provide static
 * content as well as form elements.
 *
 * To use this class, you need to declare that your class implements the 
 * SSHelper.Callbacks interface and implement four methods in your class as follows
 * (assuming that you have an NSDictionary of WOAssociations called "associations"
 * as an instance variable):<br>
 * <br><blockquote><tt>
 * public void appendToResponse( WOResponse r, WOContext c ) {<br>
 * <blockquote>
 *     SSHelper.performAppendToResponse( r, c, associations, this );<br>
 * </blockquote>
 * }<br>
 * <br>
 * public void takeValuesFromRequest( WORequest r, WOContext c ) {<br>
 * <blockquote>
 *     SSHelper.performTakeValuesFromRequest( r, c, associations, this );<br>
 * </blockquote>
 * }<br>
 * <br>
 * public void callSuperAppendToResponse( WOResponse r, WOContext c ) {<br>
 * <blockquote>
 *     super.appendToResponse( r, c );<br>
 * </blockquote>
 * }<br>
 * <br>
 * public void callSuperTakeValuesFromRequest( WORequest r, WOContext c ) {<br>
 * <blockquote>
 *     super.takeValuesFromRequest( r, c );<br>
 * </blockquote>
 * }<br>
 *</tt></blockquote>
 */
public class SSHelper extends Object {

    /**
        Used to provide callbacks into the dynamic elements so that the 
        helper can call superclass methods. 
    */
    public interface Callbacks {
	/** 
	 * Used to call the superclass appendToResponse() method from within the helper.
	 * 
	 * @param r The current WOResponse
	 * @param c The current WOContext
	 */
        public void callSuperAppendToResponse( WOResponse r, WOContext c );
	/** 
	 * Used to call the superclass takeVaulesFromResponse() method from within the helper.
	 * 
	 * @param r The current WORequest
	 * @param c The current WOContext
	 */
        public void callSuperTakeValuesFromRequest( WORequest r, WOContext c );
    }

    /** 
     * For lists of items, such as from WOCheckboxMatrix or WOBrowser, the default item
     * separator is ", ".
     */
    protected static final String DEFAULT_SEPARATOR = ", ";

    /** 
     * Cleans up the dictionary of associations to remove any that are specific to the SwitchableString framework. A WODynamicElement will take any binding that it doesn't recognize and add it to the dynamic element's tag in the HTML. This method strips out the extra bindings that are used only by the SwitchableString framework elements before the dictionary of WOAssociations is passed on to the superclass constructor.
     * 
     * @param incomingAssociations Dictionary of all incoming WOAssociations
     * @param myAssociations Array of associations that should be      
     *      removed
     * @return Dictionary of associations with myAssociations removed
     */
    public static NSDictionary cleanUpAssociations( NSDictionary incomingAssociations, NSArray myAssociations ) {

        NSMutableDictionary result = incomingAssociations.mutableClone();

        Enumeration e = myAssociations.objectEnumerator();
        while ( e.hasMoreElements() ) {
            result.removeObjectForKey( (String) e.nextElement() );
        }
        
        return result;
    }
    
    // put this here to keep it out of the javadoc and keep people
    // from trying to instantiate one of these objects. 
    private SSHelper() {
        super();
    }
    
    /** 
     * Utility method that makes it neater to pull out an association's value and
     * handles the absence of a binding gracefully.
     * 
     * @param assocName Name of the association
     * @param assocDict Dictionary of WOAssociations
     * @param c WOContext in which to evaluate the association
     * @return Value of the association in the specified WOContext
     */
    protected static Object valueForAssociation( String assocName, NSDictionary assocDict, WOContext c ) {
        
        WOAssociation oneAssoc = (WOAssociation) assocDict.objectForKey( assocName );
        return oneAssoc == null ? null :oneAssoc.valueInComponent( c.component() );

    }
    
    
    /** 
     * Boolean version of valueForAssociation().
     * 
     * @param assocName Name of the association
     * @param assocDict Dictionary of WOAssociations
     * @param c WOContext in which to evaluate the association
     * @return Boolean value of the association in the specified context
     */
    protected static boolean booleanValueForAssociation( String assocName, NSDictionary assocDict, WOContext c ) {
        
        WOAssociation oneAssoc = (WOAssociation) assocDict.objectForKey( assocName );
        return oneAssoc == null ? false : oneAssoc.booleanValueInComponent( c.component() );
        
    }

    /** 
     * Performs the appendToResponse phase for a SwitchableString element. If the
     * <CODE>editable</CODE> binding is absent, or is present and has a value of true,
     * then the element is a form element and control flow is passed on to the superclass's
     * method to generate the appropriate HTML. <BR>
     * <BR>
     * Otherwise, the content of the element is read, the content is truncated if necessary,
     * the truncated content is filtered for allowed and disallowed HTML, and the filtered
     * result is appended to the WOResponse.
     * 
     * @param r Current WOResponse
     * @param c Current WOContext
     * @param associations Dictionary of WOAssociations
     * @param dynelt Original WODynamicElement, so that superclass     
     *      functionality can be invoked
     */
    public static void performAppendToResponse( WOResponse r, WOContext c, NSDictionary associations, Callbacks dynelt ) {

        if ( associations.objectForKey( "editable" ) != null && !booleanValueForAssociation( "editable", associations, c ) ) {

            // Four possible sources of values:
            //	value
            //	selection
            //	selections
            //	noSelectionString
            
            String rawValueString = readRawValue( associations, c );
                        
            String truncatedString = truncateRawValueString( rawValueString, associations, c );
            
            String filteredContentString = filterContent( truncatedString, associations, c );
            
            r.appendContentString( filteredContentString );
            
            return;
        }
        
        // if editable is true or missing then call super
        dynelt.callSuperAppendToResponse( r, c );

    }


    /** 
     * Returns the contents of the form element, based on which bindings exist. The
     * value binding has first priority, then the selection binding, then the selections
     * binding, then the noSelectionString binding. If none of these returns any content,
     * the raw value is returned as the empty string.
     * 
     * @param associations Dictionary of WOAssociations
     * @param c The current WOContext
     * @return Raw value of the dynamic element
     */
    protected static String readRawValue( NSDictionary associations, WOContext c ) {
        
        String rawValueString; 
        Object valueObj;
        
        //put the value of the data in as a string
        // unless the value is null, in which case don't put in anything
        if ( (valueObj = valueForAssociation( "value", associations, c )) != null ) {
            
            rawValueString = stringForValue( valueObj, associations, c );
            
        } else if ( (valueObj = valueForAssociation( "selection", associations, c )) != null ) {
            
            rawValueString = stringForSelection( valueObj, associations, c );
            
        } else if ( (valueObj = valueForAssociation( "selections", associations, c )) != null ) {
            
            rawValueString = stringForSelections( (NSArray) valueObj, associations, c );
            
        } else if ( (valueObj = valueForAssociation( "noSelectionString", associations, c )) != null ) {
            
            rawValueString = valueObj.toString();
            
        } else
            
            // otherwise all of the possible bindings were null, so rawValueString becomes the empty string
            rawValueString = "";
        
        return rawValueString;
    }        
    

    /** 
     * Returns a raw value string based on the raw value passed through any formatter
     * or formatter subclass that might be bound to the dynamic element. This includes
     * formatter produced by NumberFormat and DateFormat strings.
     * 
     * @param rawValue Object returned from the value binding
     * @param associations Dictionary of WOAssociations
     * @param c The current WOContext
     * @return String from the value binding
     */
    protected static String stringForValue( Object rawValue, NSDictionary associations, WOContext c ) {

        // may have no formatter
        // may have formatter
        // may have numberformat
        // may have dateformat
        Format formatter;
        String formatString;
        
        if ( rawValue == null ) {
            
            return "";
            
        } else if ( (formatter = (Format) valueForAssociation( "formatter", associations, c )) != null ) {

            return formatter.format( rawValue );

        } else if ( (formatString = (String) valueForAssociation( "numberformat", associations, c )) != null ) {

            NSNumberFormatter nf = new NSNumberFormatter( formatString );
            return nf.format( rawValue );

        } else if ( (formatString = (String) valueForAssociation( "dateformat", associations, c )) != null ) {

            NSTimestampFormatter df = new NSTimestampFormatter( formatString );
            return df.format( rawValue );

        } else 
            
            return rawValue.toString();
            
    }



    /** 
     * Returns a raw value string based on the selection binding. We need to push
     * the selection up into the item association, then pull down the
     * displayString association, both because we need to object correctly formatted
     * for display and also because the developer might be playing tricks with
     * the accessors.
     * 
     * @param rawValue Object returned from the selection binding
     * @param associations Dictionary of WOAssociations
     * @param c The current WOContext
     * @return String from the selection binding
     */
        // need 
    protected static String stringForSelection( Object rawValue, NSDictionary associations, WOContext c ) {

        String displayStringResult;
        Object itemAssoc = associations.objectForKey( "item" );

        if ( rawValue == null )
            return "";
        if ( itemAssoc == null )
            return rawValue.toString();
        
        ((WOAssociation) itemAssoc).setValue( rawValue, c.component() );
        displayStringResult = (String) valueForAssociation( "displayString", associations, c );
        return displayStringResult == null ? rawValue.toString() : displayStringResult;
        
    }

    
    /** 
     * Returns a raw value string based on the selections binding. By calling
     * stringForSelection as a part of the process, we get the same benefits of
     * getting a properly formatted object and proper usage of accessors.
     * 
     * @param selections Object returned from the selections binding
     * @param associations Dictionary of WOAssociations
     * @param c The current WOContext
     * @return String from the selections binding
     */
    protected static String stringForSelections( NSArray selections, NSDictionary associations, WOContext c ) {
                
        if ( (selections == null) || (selections.count() == 0) ) 
            return "";
        
        Enumeration e = selections.objectEnumerator();
        StringBuffer rawValueStringBuffer = new StringBuffer();
        Object separatorObj;
        String separator = (separatorObj = valueForAssociation( "separator", associations, c )) == null ? DEFAULT_SEPARATOR : separatorObj.toString();
                
        // first one goes on without a separator
        if ( e.hasMoreElements() )
            rawValueStringBuffer.append( stringForSelection( e.nextElement(), associations, c ) );
        
        while ( e.hasMoreElements() ) {
            rawValueStringBuffer.append( separator );
            rawValueStringBuffer.append( stringForSelection( e.nextElement(), associations, c ) );
        }
        return rawValueStringBuffer.toString();
    }


    /** 
     * Truncates the raw value string at the right length. If there is no truncateLength
     * binding, then no truncation happens.
     * 
     * @param rawValueString String value from the appropriate binding for the
     *      current element
     * @param associations Dictionary of WOAssociations
     * @param c The current WOContext
     * @return Truncated string
     */
    protected static String truncateRawValueString( String rawValueString, NSDictionary associations, WOContext c ) {
        
        Number truncateLength = (Number) valueForAssociation( "truncateLength", associations, c );
        
        // truncate before filtering, so that all tags can be handled correctly
        if ( truncateLength == null )
            return rawValueString;
        
        int truncLen = truncateLength.intValue();
        
        if ( truncLen < rawValueString.length() )
            return rawValueString.substring( 0, truncLen );
        else
            return rawValueString;
    }

    
    /** 
     * Filters the HTML content based on the escapeHTML, allowedTags, and disallowedTags
     * bindings.
     * <OL>
     * <LI>If escapeHTML is true, all HTML tags are escaped.</LI>
     * <LI>If escapeHTML is false, no HTML tags are escaped.</LI>
     * <LI>If the escapeHTML, allowedTags and disallowedTags bindings are all null
     * then all tags are escaped.</LI>
     * <LI>If escapeHTML is not bound, use the allowedTags and disallowedTags bindings.</LI>
     * </OL>
     * 
     * @param truncatedString String to be filtered
     * @param associations Dictionary of WOAssociations
     * @param c The current WOContext
     * @return The value string filtered of unwanted HTML tags
     */
    protected static String filterContent( String truncatedString, NSDictionary associations, WOContext c ) {
        
        String allowedTagsString = (String) valueForAssociation( "allowedTags", associations, c );
        String disallowedTagsString = (String) valueForAssociation( "disallowedTags", associations, c );
        boolean escapeHTMLValue = booleanValueForAssociation( "escapeHTML", associations, c );
        
        if ( escapeHTMLValue )
            
            // if the escapeHTML binding exists and is true, then it overrides the
            // allowedTags and disallowedTags bindings
            return (SSHTMLFilter.Factory.filterForStrings( "", "" )).filterForAllowedHTML( truncatedString );
        
        else if ( associations.objectForKey( "escapeHTML" ) != null && !escapeHTMLValue )
            
            // if the escapeHTML binding exists and is false, then it overrides the
            // allowedTags and disallowedTags bindings and no filtering happens
            return truncatedString;
        
        else
            
            // filter on the allowed and disallowed tags
            return SSHTMLFilter.Factory.filterForStrings( allowedTagsString, disallowedTagsString ).filterForAllowedHTML( truncatedString );
    }
    
    
    /** 
     * Performs the takeValuesFromRequest phase for a SwitchableString element. If the
     * <CODE>editable</CODE> binding is absent, or is present and has a value of true,
     * then the element is a form element and control flow is passed on to the superclass's
     * method to extract the form values from the WORequest. <BR>
     * <BR>
     * Otherwise, the takeValue phase is skipped since we don't want to read in a bogus
     * null value.
     * 
     * @param r The current WORequest
     * @param c The current WOContext
     * @param associations Dictionary of WOAssociations
     * @param dynelt Original WODynamicElement, so that superclass 
     *     functionality can be invoked
     */
    public static void performTakeValuesFromRequest( WORequest r, WOContext c, NSDictionary associations, Callbacks dynelt ) {
        // if editable is false then do nothing, since we don't want the association to
        // push a null value up to the parent.
        if ( associations.objectForKey( "editable" ) != null ) {
            if ( !((WOAssociation) associations.objectForKey( "editable" )).booleanValueInComponent( c.component() ) )
                return;
        }
        // if editable is true or missing then call super
        dynelt.callSuperTakeValuesFromRequest( r, c );

    }

}
